import json
from abc import ABC, abstractmethod
from typing import Optional
from urllib.request import Request, urlopen

import yapper.constants as c
from yapper.enums import GeminiModel, GroqModel, Persona


def enhancer_gemini(
    model: str, persona_instr: str, api_key: str, text: str
) -> str:
    """
    Enhances the given text using Gemini.

    Parameters
    ----------
    model : str
        The gemini model to use for enhancing text.
    persona_instr : the persona instruction to use for text enhancement,
        can be any system message.
    api_key : str
        The Gemini API key used for making request.
    query: str
        The text to enhance.
    """
    base = "https://generativelanguage.googleapis.com/v1beta/models"
    url = f"{base}/{model}:generateContent?key={api_key}"
    headers = {
        "Content-Type": "application/json",
    }
    data = {
        c.GEMINI_FLD_SYS_INST: {
            c.GEMINI_FLD_PARTS: {c.GEMINI_FLD_TEXT: persona_instr}
        },
        c.GEMINI_FLD_CONTENTS: {c.GEMINI_FLD_PARTS: {c.GEMINI_FLD_TEXT: text}},
    }
    request = Request(
        url, headers=headers, data=json.dumps(data).encode("utf-8")
    )
    with urlopen(request) as response:
        data = json.loads(response.read())
        return data[c.GEMINI_FLD_CANDIDATES][0][c.GEMINI_FLD_CONTENT][
            c.GEMINI_FLD_PARTS
        ][0][c.GEMINI_FLD_TEXT]


def enhancer_groq(
    model: str, api_key: str, persona_instr: str, text: str
) -> str:
    url = "https://api.groq.com/openai/v1/chat/completions"
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {api_key}",
        "User-Agent": "curl/7.68.0",
    }
    data = {
        c.GROQ_FLD_MESSAGES: [
            {c.FLD_ROLE: c.ROLE_SYSTEM, c.FLD_CONTENT: persona_instr},
            {c.FLD_ROLE: c.ROLE_USER, c.FLD_CONTENT: text},
        ],
        c.GROQ_FLD_MODEL: model,
    }
    with urlopen(
        Request(url, headers=headers, data=json.dumps(data).encode("utf-8"))
    ) as response:
        data = json.loads(response.read())
        return data[c.FLD_CHOICES][0][c.FLD_MESSAGE][c.FLD_CONTENT]


class BaseEnhancer(ABC):
    """
    Base class for text enhancers.

    Methods
    ----------
    enhance(text: str) -> str
        Enhances the given text.
    """

    @abstractmethod
    def enhance(self, text: str) -> str:
        pass


class NoEnhancer(BaseEnhancer):
    """Doesn't do anything to the text, a placeholder for a real enhancer."""

    def enhance(self, text: str):
        return text


class GeminiEnhancer(BaseEnhancer):
    """Enhances text using a Gemini model."""

    def __init__(
        self,
        api_key: str,
        model: GeminiModel = GeminiModel.GEMINI_1_5_PRO,
        persona: Persona = Persona.DEFAULT,
        persona_instr: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        api_key : str
            Your gemini api key.
        gemini_model : GeminiModel, optional
            the gemini model to use for enhancement, must be one of
            'GeminiModel' enum's attributes
            (default: GeminiModel.PRO_1_5_002).
        persona : Persona, optional
            The persona to be used for enhancement (default: Persona.DEFAULT).
        persona_instr : Optional[str]
            Custom persona instruction, can be used to give the LLM a custom
            persona (default: None).
        """
        if persona_instr is not None:
            self.persona_instr = persona_instr
        else:
            assert (
                persona in Persona
            ), f"persona must be one of {', '.join(Persona)}"
            self.persona_instr = c.persona_instrs[persona]
        self.model = model.value
        self.api_key = api_key
        self.default_enhancer = None

    def enhance(self, text: str) -> str:
        """Return text enhanced by Groq API."""
        return enhancer_gemini(
            self.model, self.persona_instr, self.api_key, text
        )


class GroqEnhancer(BaseEnhancer):
    """
    Enhances text using a Groq API
    (https://console.groq.com/docs/overview).
    """

    def __init__(
        self,
        api_key: str,
        model: GroqModel = GroqModel.LLAMA_3_1_8B_INSTANT,
        persona: Persona = Persona.DEFAULT,
        persona_instr: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        api_key : str
            Your Groq api key.
        groq_model : GroqModel, optional
            the model to use for enhancement, must be one of 'GroqModel'
            enum's attributes (default: GroqModel.LLAMA_3_8B_8192).
        persona : Persona, optional
            The persona to be used for enhancement (default: Persona.DEFAULT).
        persona_instr : Optional[str]
            Custom persona instruction, can be used to give the LLM a custom
            persona (default: None).
        """
        if persona_instr is not None:
            self.persona_instr = persona_instr
        else:
            assert (
                persona in Persona
            ), f"persona must be one of {', '.join(Persona)}"
            self.persona_instr = c.persona_instrs[persona]
        self.model = model
        self.api_key = api_key
        self.default_enhancer = None

    def enhance(self, text: str) -> str:
        """Returns text enhanced by Groq API."""
        return enhancer_groq(
            self.model.value, self.api_key, self.persona_instr, text
        )



import os
import platform
import random
import string
import sys
import tarfile
import unicodedata
import zipfile
from pathlib import Path
from urllib.parse import quote
from urllib.request import urlretrieve

import yapper.constants as c
import yapper.meta as meta
from yapper.enums import PiperQuality, PiperVoice

PLATFORM = None
APP_DIR = None

if os.name == "nt":
    PLATFORM = c.PLATFORM_WINDOWS
    APP_DIR = Path(os.getenv("APPDATA"))
elif os.name == "posix":
    home = Path.home()
    if os.uname().sysname == "Darwin":
        PLATFORM = c.PLATFORM_MAC
        APP_DIR = Path.home() / "Library/Application Support"
    else:
        PLATFORM = c.PLATFORM_LINUX
        APP_DIR = Path.home() / ".config"
else:
    print("your system is not supported")
    sys.exit()

APP_DIR = APP_DIR / meta.name
APP_DIR.mkdir(exist_ok=True)


def get_random_name(length: int = 10) -> str:
    """
    Creates a 'length' letter random string.

    Parameters
    ----------
    length : int, optional
        Length of the random string (default: 10).
    """
    return "".join(random.choices(string.ascii_letters, k=length))


def normalize_path(name: str) -> Path:
    norm = (
        unicodedata.normalize("NFKD", str(name))
        .encode("ascii", "ignore")
        .decode("utf-8")
    )
    return Path(norm)


def progress_hook(block_idx: int, block_size: int, total_bytes: int):
    """Shows download progress."""
    part = min(((block_idx + 1) * block_size) / total_bytes, 1)
    progress = "=" * int(60 * part)
    padding = " " * (60 - len(progress))
    print("\r[" + progress + padding + "]", end="")


def download(url: str, file: str, show_progress: bool):
    """
    Downloads the content from the given URL into the given file.

    Parameters
    ----------
    url : str
        The URL to download content from.
    file : str
        The file to save the URL content into.
    show_progress: bool, optional
        Whether to show progress while downloading.
    """
    hook = progress_hook if show_progress else None
    urlretrieve(url, file, reporthook=hook)
    if show_progress:
        print("")


def install_piper(show_progress: bool) -> Path:
    """Installs piper into the app's home directory."""
    exe_path = (
        APP_DIR
        / "piper"
        / ("piper.exe" if PLATFORM == c.PLATFORM_WINDOWS else "piper")
    )
    marker_file = APP_DIR / "piper_installed"
    if marker_file.exists():
        return exe_path
    zip_path = APP_DIR / "piper.zip"
    if show_progress:
        print("installing piper...")
    prefix = "https://github.com/rhasspy/piper/releases/download/2023.11.14-2"
    if PLATFORM == c.PLATFORM_LINUX:
        if platform.machine() in ("aarch64", "arm64"):
            nix_link = f"{prefix}/piper_linux_aarch64.tar.gz"
        elif platform.machine() in ("armv7l", "armv7"):
            nix_link = f"{prefix}/piper_linux_armv7l.tar.gz"
        else:
            nix_link = f"{prefix}/piper_linux_x86_64.tar.gz"
        download(nix_link, zip_path, show_progress)
    elif PLATFORM == c.PLATFORM_WINDOWS:
        download(f"{prefix}/piper_windows_amd64.zip", zip_path, show_progress)
    else:
        download(f"{prefix}/piper_macos_x64.tar.gz", zip_path, show_progress)

    if PLATFORM == c.PLATFORM_WINDOWS:
        with zipfile.ZipFile(zip_path, "r") as z_f:
            z_f.extractall(APP_DIR)
    else:
        with tarfile.open(zip_path, "r") as z_f:
            z_f.extractall(APP_DIR)
    os.remove(zip_path)
    marker_file.write_bytes(b"")
    return exe_path


def download_piper_model(
    voice: PiperVoice,
    quality: PiperQuality,
    show_progress: bool,
) -> tuple[Path, Path]:
    """
    Downloads the given piper voice with the given quality.

    Parameters
    ----------
    voice : PiperVoiceUS or PiperVoiceGB
        The Piper voice model to download.
    quality : PiperQuality
        The quality of the given voice.

    Returns
    ----------
    tuple of str
        The voice model file and the voice configuration file in a tuple.
    """
    voices_dir = APP_DIR / "piper_voices"
    voices_dir.mkdir(exist_ok=True)
    lang_code = c.piper_enum_to_lang_code[voice.__class__]
    voice, quality = voice.value, quality.value

    marker_file = voices_dir / f"{lang_code}-{voice}-{quality}"
    onnx_file = voices_dir / f"{lang_code}-{voice}-{quality}.onnx"
    conf_file = voices_dir / f"{lang_code}-{voice}-{quality}.onnx.json"

    prefix = "https://huggingface.co/rhasspy/piper-voices/resolve/main/"
    prefix += lang_code.split("_")[0] + "/" + lang_code

    onnx_url = f"{prefix}/{voice}/{quality}/{onnx_file.name}?download=true"
    conf_url = f"{prefix}/{voice}/{quality}/{conf_file.name}?download=true"

    marker_file = normalize_path(marker_file)
    onnx_file = normalize_path(onnx_file)
    conf_file = normalize_path(conf_file)

    if marker_file.exists():
        return onnx_file, conf_file
    if not onnx_file.exists():
        try:
            if show_progress:
                print(f"downloading requirements for {voice}...")
            download(quote(onnx_url, safe=":/?=&"), onnx_file, show_progress)
        except (KeyboardInterrupt, Exception) as e:
            onnx_file.unlink(missing_ok=True)
            if getattr(e, "status", None) == 404:
                raise Exception(
                    f"{voice}({quality}) is not available, please refer to"
                    f" {prefix} to check all available models"
                )
            raise e
    if not conf_file.exists():
        try:
            download(quote(conf_url, safe=":/?=&"), conf_file, show_progress)
        except (KeyboardInterrupt, Exception) as e:
            conf_file.unlink(missing_ok=True)
            raise e
    marker_file.write_bytes(b"")
    return onnx_file, conf_file



from enum import Enum


class PiperVoice(str, Enum):
    pass


class PiperVoiceJordan(PiperVoice):
    KAREEM = "kareem"


class PiperVoiceCatalonia(PiperVoice):
    UPC_ONA = "upc_ona"
    UPC_PAU = "upc_pau"


class PiperVoiceCzech(PiperVoice):
    JIRKA = "jirka"


class PiperVoiceWales(PiperVoice):
    BU_TTS = "bu_tts"
    GWRYW_GOGLEDDOL = "gwryw_gogleddol"


class PiperVoiceDenmark(PiperVoice):
    TALESYNTESE = "talesyntese"


class PiperVoiceGermany(PiperVoice):
    EVA_K = "eva_k"
    KARLSSON = "karlsson"
    KERSTIN = "kerstin"
    MLS = "mls"
    PAVOQUE = "pavoque"
    RAMONA = "ramona"
    THORSTEN = "thorsten"
    THORSTEN_EMOTIONAL = "thorsten_emotional"


class PiperVoiceGreece(PiperVoice):
    RAPUNZELINA = "rapunzelina"


class PiperVoiceSpain(PiperVoice):
    CARLFM = "carlfm"
    DAVEFX = "davefx"
    MLS_10246 = "mls_10246"
    MLS_9972 = "mls_9972"
    SHARVARD = "sharvard"


class PiperVoiceMexico(PiperVoice):
    ALD = "ald"
    CLAUDE = "claude"


class PiperVoiceIran(PiperVoice):
    AMIR = "amir"
    GANJI = "ganji"
    GANJI_ADABI = "ganji_adabi"
    GYRO = "gyro"
    REZA_IBRAHIM = "reza_ibrahim"


class PiperVoiceFinland(PiperVoice):
    HARRI = "harri"


class PiperVoiceFrance(PiperVoice):
    GILLES = "gilles"
    MLS = "mls"
    MLS_1840 = "mls_1840"
    SIWIS = "siwis"
    TOM = "tom"
    UPMC = "upmc"


class PiperVoiceHungary(PiperVoice):
    ANNA = "anna"
    BERTA = "berta"
    IMRE = "imre"


class PiperVoiceIceland(PiperVoice):
    BUI = "bui"
    SALKA = "salka"
    STEINN = "steinn"
    UGLA = "ugla"


class PiperVoiceItaly(PiperVoice):
    PAOLA = "paola"
    RICCARDO = "riccardo"


class PiperVoiceGeorgia(PiperVoice):
    NATIA = "natia"


class PiperVoiceKazakhstan(PiperVoice):
    ISEKE = "iseke"
    ISSAI = "issai"
    RAYA = "raya"


class PiperVoiceLuxembourg(PiperVoice):
    MARYLUX = "marylux"


class PiperVoiceLatvia(PiperVoice):
    AIVARS = "aivars"


class PiperVoiceNepal(PiperVoice):
    GOOGLE = "google"


class PiperVoiceBelgium(PiperVoice):
    NATHALIE = "nathalie"
    RDH = "rdh"


class PiperVoiceNetherlands(PiperVoice):
    MLS = "mls"
    MLS_5809 = "mls_5809"
    MLS_7432 = "mls_7432"
    PIM = "pim"
    RONNIE = "ronnie"


class PiperVoiceNorway(PiperVoice):
    TALESYNTESE = "talesyntese"


class PiperVoicePoland(PiperVoice):
    DARKMAN = "darkman"
    GOSIA = "gosia"
    MC_SPEECH = "mc_speech"


class PiperVoiceBrazil(PiperVoice):
    CADU = "cadu"
    EDRESSON = "edresson"
    FABER = "faber"
    JEFF = "jeff"


class PiperVoicePortugal(PiperVoice):
    TUGAO = "tugão"


class PiperVoiceRomania(PiperVoice):
    MIHAI = "mihai"


class PiperVoiceRussia(PiperVoice):
    DENIS = "denis"
    DMITRI = "dmitri"
    IRINA = "irina"
    RUSLAN = "ruslan"


class PiperVoiceSlovakia(PiperVoice):
    LILI = "lili"


class PiperVoiceSlovenia(PiperVoice):
    ARTUR = "artur"


class PiperVoiceSerbia(PiperVoice):
    SERBSKI_INSTITUT = "serbski_institut"


class PiperVoiceSweden(PiperVoice):
    LISA = "lisa"
    NST = "nst"


class PiperVoiceCongo(PiperVoice):
    LANFRICA = "lanfrica"


class PiperVoiceTurkey(PiperVoice):
    DFKI = "dfki"
    FAHRETTIN = "fahrettin"
    FETTAH = "fettah"


class PiperVoiceUkraine(PiperVoice):
    LADA = "lada"
    UKRAINIAN_TTS = "ukrainian_tts"


class PiperVoiceVietnam(PiperVoice):
    HOURS_SINGLE = "25hours_single"
    VAIS1000 = "vais1000"
    VIVOS = "vivos"


class PiperVoiceChina(PiperVoice):
    HUAYAN = "huayan"


class PiperVoiceUS(PiperVoice):
    AMY = "amy"
    ARCTIC = "arctic"
    BRYCE = "bryce"
    JOHN = "john"
    NORMAN = "norman"
    DANNY = "danny"
    HFC_FEMALE = "hfc_female"
    HFC_MALE = "hfc_male"
    JOE = "joe"
    KATHLEEN = "kathleen"
    KRISTIN = "kristin"
    LJSPEECH = "ljspeech"
    KUSAL = "kusal"
    L2ARCTIC = "l2arctic"
    LESSAC = "lessac"
    LIBRITTS = "libritts"
    LIBRITTS_R = "libritts_r"
    RYAN = "ryan"
    REZA_IBRAHIM = "reza_ibrahim"
    SAM = "sam"


class PiperVoiceGB(PiperVoice):
    ALAN = "alan"
    ALBA = "alba"
    ARU = "aru"
    CORI = "cori"
    JENNY_DIOCO = "jenny_dioco"
    NORTHERN_ENGLISH_MALE = "northern_english_male"
    SEMAINE = "semaine"
    SOUTHERN_ENGLISH_FEMALE = "southern_english_female"
    VCTK = "vctk"


class PiperVoice(Enum):
    JORDAN = PiperVoiceJordan
    CATALONIA = PiperVoiceCatalonia
    CZECH = PiperVoiceCzech
    WALES = PiperVoiceWales
    DENMARK = PiperVoiceDenmark
    GERMANY = PiperVoiceGermany
    GREECE = PiperVoiceGreece
    SPAIN = PiperVoiceSpain
    MEXICO = PiperVoiceMexico
    IRAN = PiperVoiceIran
    FINLAND = PiperVoiceFinland
    FRANCE = PiperVoiceFrance
    HUNGARY = PiperVoiceHungary
    ICELAND = PiperVoiceIceland
    ITALY = PiperVoiceItaly
    GEORGIA = PiperVoiceGeorgia
    KAZAKHSTAN = PiperVoiceKazakhstan
    LUXEMBOURG = PiperVoiceLuxembourg
    LATVIA = PiperVoiceLatvia
    NEPAL = PiperVoiceNepal
    BELGIUM = PiperVoiceBelgium
    NETHERLANDS = PiperVoiceNetherlands
    NORWAY = PiperVoiceNorway
    POLAND = PiperVoicePoland
    BRAZIL = PiperVoiceBrazil
    PORTUGAL = PiperVoicePortugal
    ROMANIA = PiperVoiceRomania
    RUSSIA = PiperVoiceRussia
    SLOVAKIA = PiperVoiceSlovakia
    SLOVENIA = PiperVoiceSlovenia
    SERBIA = PiperVoiceSerbia
    SWEDEN = PiperVoiceSweden
    CONGO = PiperVoiceCongo
    TURKEY = PiperVoiceTurkey
    UKRAINE = PiperVoiceUkraine
    VIETNAM = PiperVoiceVietnam
    CHINA = PiperVoiceChina
    US = PiperVoiceUS
    GB = PiperVoiceGB


class PiperQuality(str, Enum):
    X_LOW = "x_low"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class GroqModel(str, Enum):
    GEMMA_2_9B_IT = "gemma2-9b-it"
    LLAMA_3_3_70B_VERSATILE = "llama-3.3-70b-versatile"
    LLAMA_3_1_8B_INSTANT = "llama-3.1-8b-instant"
    LLAMA_GUARD_3_8B = "llama-guard-3-8b"
    LLAMA_2_70B_8192 = "llama3-70b-8192"
    LLAMA_2_8B_8192 = "llama3-8b-8192"
    ALLAM_2_7B = "allam-2-7b"
    DEEPSEEK_R1_DISTILL_LLAMA_70B = "deepseek-r1-distill-llama-70b"
    LLAMA_4_MAVERICK_17B_128E_INSTRUCT = (
        "meta-llama/llama-4-maverick-17b-128e-instruct"
    )
    LLAMA_4_SCOUT_17B_16E_INSTRUCT = (
        "meta-llama/llama-4-scout-17b-16e-instruct"
    )
    QWEN_QWQ_32B = "qwen-qwq-32b"


class GeminiModel(str, Enum):
    GEMINI_2_5_FLASH_PREVIEW_04_17 = "gemini-2.5-flash-preview-04-17"
    GEMINI_2_0_FLASH = "gemini-2.0-flash"
    GEMINI_2_0_FLASH_LITE = "gemini-2.0-flash-lite"
    GEMINI_1_5_FLASH = "gemini-1.5-flash"
    GEMINI_1_5_FLASH_8B = "gemini-1.5-flash-8b"
    GEMINI_1_5_PRO = "gemini-1.5-pro"
    GEMINI_2_0_FLASH_THINKING_EXP_01_21 = "gemini-2.0-flash-thinking-exp-01-21"
    GEMINI_2_0_PRO_EXP_02_05 = "gemini-2.0-pro-exp-02-05"
    GEMINI_2_0_FLASH_EXP = "gemini-2.0-flash-exp"
    GEMINI_EXP_1206 = "gemini-exp-1206"
    GEMINI_2_0_FLASH_THINKING_EXP_1219 = "gemini-2.0-flash-thinking-exp-1219"
    GEMINI_1_5_FLASH_8B_EXP_0924 = "gemini-1.5-flash-8b-exp-0924"
    GEMINI_1_5_FLASH_8B_EXP_0827 = "gemini-1.5-flash-8b-exp-0827"


class Persona(str, Enum):
    DEFAULT = "companion"
    JARVIS = "jarvis"
    FRIDAY = "friday"
    ALFRED = "alfred"
    HAL = "HAL"
    CORTANA = "cortana"
    SAMANTHA = "samantha"
    TARS = "TARS"



# Copyright 2025 Arvin Adeli
# Licensed under the Apache License, Version 2.0
# http://www.apache.org/licenses/LICENSE-2.0

def main():
    import os
    os.environ["VOSK_LOG_LEVEL"] = "0"

    import vosk
    import pyaudio
    import json
    import subprocess
    from collections import deque
    import requests
    from jarvis_os.yapper.core import Yapper, PiperSpeaker
    from jarvis_os.yapper.enhancer import GroqEnhancer
    from jarvis_os.yapper.speaker import BaseSpeaker
    from jarvis_os.yapper.speaker import PiperVoiceUS
    import threading
    import queue
    import urllib.request
    import zipfile
    import sys

    state = {
        "awaiting_shutdown" : False,
        "is_speaking" : False,
        "is_stopped" : False

    }

    from rich.console import Console
    from rich.live import Live
    from rich.panel import Panel
    from rich.text import Text
    import time
    from rich.progress import Progress, BarColumn, TextColumn, TimeElapsedColumn
    from rich.progress import SpinnerColumn
    import itertools
    import contextlib


    def ensure_vosk_model():
        base_dir = os.path.expanduser("~/.jarvis-os/models")
        model_name = "vosk-model-en-us-0.42-gigaspeech"
        model_path = os.path.join(base_dir, model_name)

        if not os.path.exists(model_path):
            print("Vosk model not found. Downloading...")
            os.makedirs(base_dir, exist_ok=True)

            zip_path = os.path.join(base_dir, f"{model_name}.zip")
            url = f"https://alphacephei.com/vosk/models/{model_name}.zip"
            urllib.request.urlretrieve(url, zip_path)

            print("Extracting model...")
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                zip_ref.extractall(base_dir)
            os.remove(zip_path)
            print("Model setup complete.")

        return model_path
    
    def glowing_boot_animation():

        ascii_jarvis = [
                "    _   _    ______     _____ ____         ___  ____  ",
                "     | | / \  |  _ \ \   / /_ _/ ___|       / _ \/ ___| ",
                "  _  | |/ _ \ | |_) \ \ / / | |\___ \ _____| | | \___ \ ",
                " | |_| / ___ \|  _ < \ V /  | | ___) |_____| |_| |___) |",
                "  \___/_/   \_\_| \_\ \_/  |___|____/       \___/|____/ ",
            ]

        with Live(refresh_per_second=5) as live:
            for i in range(1, len(ascii_jarvis) + 1):
                lines = ascii_jarvis[:i]
                styled_text = Text.from_markup("\n".join(f"[bold bright_magenta]{line}" for line in lines), justify="center")
                panel = Panel(styled_text, border_style="bold magenta", padding=(1, 4))
                live.update(panel)
                time.sleep(0.4)

            # Final glow
            time.sleep(0.3)
            final_text = Text.from_markup("\n".join(f"[bold cyan]{line}" for line in ascii_jarvis), justify="center")
            panel = Panel(final_text, border_style="bold blue", padding=(1, 4))
            live.update(panel)
            time.sleep(1.2)

    @contextlib.contextmanager
    def suppress_stderr(log_path = "~/.jarvis-os/logs/jarvis.log"):
        log_path = os.path.expanduser(log_path)
        os.makedirs(os.path.dirname(log_path), exist_ok=True)

        with open(log_path, 'w') as log_file:
            stderr_fileno = sys.stderr.fileno()
            with os.fdopen(os.dup(stderr_fileno), 'w') as old_stderr:
                os.dup2(log_file.fileno(), stderr_fileno)
                try:
                    yield
                finally:
                    os.dup2(old_stderr.fileno(), stderr_fileno)

    def load_vosk_model_with_progress():
        model_path = ensure_vosk_model()

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TimeElapsedColumn(),
            transient=True,
        ) as progress:
            task = progress.add_task("[bold bright_magenta]Loading Vosk Speech Recognition Model...", total=100)
            for _ in range(20):
                time.sleep(0.05)
                progress.update(task, advance=5)
            with suppress_stderr("~/.jarvis-os/logs/jarvis.log"):
                model = vosk.Model(model_path)
        return model

    def load_vosk_model_with_spinner():

        model_path = ensure_vosk_model()
        #spinner_cycle = itertools.cycle(["|  |  |  |  |  |", "J  A  R  V  I  S", "|  |  |  |  |  |", "/  /  /  /  /  /", "-  -  -  -  -  -", "\\  \\  \\  \\  \\  \\"])
        spinner_cycle = itertools.cycle([
                                            "|  |  |  |  |  |",       # classic spinner
                                            "/  /  /  /  /  /",
                                            "-  -  -  -  -  -",
                                            "\\  \\  \\  \\  \\  \\",

                                            "J  |  |  |  |  |",       # morph begins
                                            "J  A  |  |  |  |",
                                            "J  A  R  |  |  |",
                                            "J  A  R  V  |  |",
                                            "J  A  R  V  I  |",
                                            "J  A  R  V  I  S",       # full reveal
                                            "J  A  R  V  I  S",
                                            "J  A  R  V  I  |",       # morph out
                                            "J  A  R  V  |  |",
                                            "J  A  R  |  |  |",
                                            "J  A  |  |  |  |",
                                            "J  |  |  |  |  |",
                                            "|  |  |  |  |  |",       # back to spinner
                                        ])
        loading_text = "Loading Vosk Speech Recognition Model:  "

        with Live(refresh_per_second=10) as live:
            start_time = time.time()
            done = threading.Event()

            def animate_spinner():
                while not done.is_set():
                    spinner = next(spinner_cycle)
                    text = Text(f"{loading_text} {spinner}", style="bold green")
                    live.update(Panel(text, border_style="green"))
                    time.sleep(0.1)

            spinner_thread = threading.Thread(target=animate_spinner, daemon=True)
            spinner_thread.start()

            # Load model (this part blocks the main thread)
            with suppress_stderr("~/.jarvis-os/logs/jarvis.log"):
                model = vosk.Model(model_path)

            done.set()  # stop spinner
            spinner_thread.join()

            # Show success panel
            elapsed = time.time() - start_time
            success = Text(f"✓ Model loaded in {elapsed:.1f} seconds", style="bold cyan")
            live.update(Panel(success, border_style="cyan"))
            time.sleep(1)

        return model



    glowing_boot_animation()

    # def glowing_boot_animation():
    #     console = Console()

    #     ascii_jarvis = [
    #         "    _   _    ______     _____ ____         ___  ____  ",
    #         "     | | / \  |  _ \ \   / /_ _/ ___|       / _ \/ ___| ",
    #         "  _  | |/ _ \ | |_) \ \ / / | |\___ \ _____| | | \___ \ ",
    #         " | |_| / ___ \|  _ < \ V /  | | ___) |_____| |_| |___) |",
    #         "  \___/_/   \_\_| \_\ \_/  |___|____/       \___/|____/ ",
    #     ]

    #     with Live(refresh_per_second=5) as live:
    #         for i in range(1, len(ascii_jarvis) + 1):
    #             lines = ascii_jarvis[:i]
    #             styled_text = Text.from_markup("\n".join(f"[bold cyan]{line}" for line in lines), justify="center")
    #             panel = Panel(styled_text, border_style="bold blue", padding=(1, 4))
    #             live.update(panel)
    #             time.sleep(0.4)

    #         # Final flash or glow
    #         time.sleep(0.3)
    #         final_text = Text.from_markup("\n".join(f"[bold bright_magenta]{line}" for line in ascii_jarvis), justify="center")
    #         panel = Panel(final_text, border_style="bold magenta", padding=(1, 4))
    #         live.update(panel)
    #         time.sleep(1.2)




    # @contextlib.contextmanager
    # def suppress_stderr():
    #     with open(os.devnull, 'w') as fnull:
    #         stderr_fileno = sys.stderr.fileno()
    #         with os.fdopen(os.dup(stderr_fileno), 'w') as old_stderr:
    #             os.dup2(fnull.fileno(), stderr_fileno)
    #             try:
    #                 yield
    #             finally:
    #                 os.dup2(old_stderr.fileno(), stderr_fileno)

    # def load_vosk_model_with_progress():
    #     model_path = ensure_vosk_model()

    #     with Progress(
    #         SpinnerColumn(),
    #         TextColumn("[progress.description]{task.description}"),
    #         BarColumn(),
    #         TimeElapsedColumn(),
    #         transient=True,
    #     ) as progress:
    #         task = progress.add_task("[green]Loading Vosk Speech Recognition Model...", total=100)

    #         for _ in range(20):
    #             time.sleep(0.05)
    #             progress.update(task, advance=5)

    #         # Suppress Vosk logs while loading the model
    #         with suppress_stderr():
    #             model = vosk.Model(model_path)

    #     return model

    model = load_vosk_model_with_spinner()

    #model = vosk.Model(r"C:\Users\arvin\Documents\Jarvis\vosk-model-en-us-0.42-gigaspeech")
    #engine = pyttsx3.init()
    lessac = PiperSpeaker(voice = PiperVoiceUS.JOE)
    yapr = Yapper(speaker = lessac)

    sentence_queue = queue.Queue()
    stop_event = threading.Event()


    def yap_thread():
        while True:
            sentence = sentence_queue.get()
            if stop_event.is_set():
                # Drain the rest of the queue without speaking
                with sentence_queue.mutex:
                    sentence_queue.queue.clear()
                continue
            if sentence:
                state["is_speaking"] = True
                yapr.yap(sentence.strip())
                state["is_speaking"] = False

    threading.Thread(target=yap_thread, daemon=True).start()
    
    # Initialize microphone input
    p = pyaudio.PyAudio()
    stream = p.open(format=pyaudio.paInt16, channels=1, rate=16000, 
                    input=True, frames_per_buffer=4000)
    stream.start_stream()

    recognizer = vosk.KaldiRecognizer(model, 16000)


    def shutdown(command):
        if "shutdown" in command.lower() or "shut down" in command.lower() :
            print("Are you sure you want to shutdown now?")
            yapr.yap("Are you sure you want to shutdown now?")
            state["awaiting_shutdown"] = True
            return True
        else:
            return False


    def google(command):
        if "google" in command.lower() or "look up" in command.lower():
            search = command.lower().replace("google", "").replace("look up", "").strip()
            words = search.split()
            url = "https://www.google.com/search?q="

            for item in words:
                if url != "https://www.google.com/search?q=":
                    url += "+"
                url += item

            try:
                subprocess.run(["start", url], shell=True, check=True)
                sentence_queue.put("Sure. Googling " + search + " now.")
            except Exception as e:
                print(f"Google Search Failed: {e}")
            return True
        else:
            return False

    def youtube(command):
        if "search youtube" in command.lower() or "youtube search" in command.lower() or "youtube search for" in command.lower():
            search = command.lower().replace("youtube search for", "").replace("youtube search", "").replace("search youtube", "").strip()
            words = search.split()
            url = "https://www.youtube.com/results?search_query="

            for item in words:
                if url != "https://www.youtube.com/results?search_query=":
                    url += "+"
                url += item

            try:
                subprocess.run(["start", url], shell=True, check=True)
                sentence_queue.put("Sure. Searching YouTube for " + search + " now.")
            except Exception as e:
                print(f"YouTube Search Failed: {e}")
            return True
        else:
            return False


    def open_software(command):
        if "open" in command.lower() or "start" in command.lower():
            app_name = command.lower().replace("open ", "").replace("start ", "").strip()
            
            # Check system PATH for executable
            def is_executable_in_path(executable):
                for path in os.environ["PATH"].split(os.pathsep):
                    full_path = os.path.join(path, executable)
                    if os.path.isfile(full_path) and os.access(full_path, os.X_OK):
                        return full_path
                return None

            # Define system-specific commands for non-executable tools
            system_commands = {
                "settings": "start ms-settings:",
                "file explorer": "explorer",
                "copilot": "start ms-copilot:"  # Assuming Copilot can be started via a custom URL scheme or command
                # Add more system-specific commands as needed
            }

            # Try to find the executable in PATH, common paths, or perform BFS search
            executable_path = is_executable_in_path(app_name) or is_executable_in_path(app_name + ".exe")
            if not executable_path:
                common_paths = [
                    os.path.expanduser("~/Desktop"),
                    os.environ.get('ProgramFiles', 'C:/Program Files'),
                    os.environ.get('ProgramFiles(x86)', 'C:/Program Files (x86)'),
                    r"C:\Users\arvin\AppData\Roaming\Microsoft\Windows\Start Menu\Programs"
                ]
                executable_path = bfs_search_executable(common_paths, app_name)
                if not executable_path:
                    # Fallback to search the entire OS starting from the root directory
                    executable_path = bfs_search_executable([os.path.abspath(os.sep)], app_name)
            
            # Execute system commands or open executables
            if executable_path:
                try:
                    subprocess.Popen([executable_path])
                    engineOutput = f"Opening {app_name}..."
                    sentence_queue.put("Got it. Opening " + app_name)
                    #print(f"Opening {app_name}...")
                    #engine.say(engineOutput)
                    #engine.runAndWait()

                except Exception as e:
                    print(f"Error opening {app_name}: {e}")
            elif app_name in system_commands:
                try:
                    subprocess.Popen([system_commands[app_name]], shell=True)
                    sentence_queue.put("Got it. Opening " + app_name)
                    print(f"Opening {app_name}...")
                except Exception as e:
                    print(f"Error opening {app_name}: {e}")
            else:
                print(f"Could not find the application: {app_name} in PATH or common search directories")
            
            return True
        else:
            return False

    def bfs_search_executable(start_dirs, app_name):
        queue = deque(start_dirs)
        visited = set()
        
        while queue:
            current_dir = queue.popleft()
            if current_dir in visited:
                continue
            visited.add(current_dir)

            try:
                with os.scandir(current_dir) as it:
                    for entry in it:
                        if entry.is_file() and entry.name.lower() == app_name.lower() + '.exe':
                            return entry.path
                        elif entry.is_dir():
                            queue.append(entry.path)
            except PermissionError:
                continue
        return None

    def ask_lm_studio_streaming(question):
        url = "https://59b5-2600-4040-2c2c-2700-b4ca-17ee-134f-b0e1.ngrok-free.app/v1/chat/completions"
        headers = {
            "Authorization": "Bearer llama-3.1-8b-lexi-uncensored-v2",
            "Content-Type": "application/json"
        }
        data = {
            "messages": [{"role": "user", "content": question}],
            "max_tokens": 5000,
            "stream": True
        }

        try:
            response = requests.post(url, headers=headers, json=data, stream=True)
            sentence = ""
            for line in response.iter_lines():
                if line:
                    decoded_line = line.decode('utf-8').strip()
                    if decoded_line.startswith("data: "):
                        decoded_line = decoded_line[len("data: "):]

                    if not decoded_line or decoded_line == "[DONE]":
                        continue

                    try:
                        json_data = json.loads(decoded_line)
                        result = (
                            json_data.get("choices", [{}])[0].get("delta", {}).get("content", "") or
                            json_data.get("choices", [{}])[0].get("text", "") or
                            json_data.get("choices", [{}])[0].get("message", {}).get("content", "")
                        )
                        if result:
                            print(result, end="", flush=True)
                            sentence += result
                            if any(result.endswith(p) for p in [".", "?", "!", ";"]):
                                sentence_queue.put(sentence)
                                sentence = ""
                                print()
                    except json.JSONDecodeError:
                        continue
        except Exception as e:
            print(f"Error: {e}")


    def docs(command):
        if "create a new document" in command.lower() or "create a new doc" in command.lower() or "create a doc" in command.lower() or "create a google doc" in command.lower():
            url = "https://docs.new"

            try:
                subprocess.run(["start", url], shell=True, check=True)
                sentence_queue.put("Sure, here's a new document.")
            except Exception as e:
                print(f"Google Docs Failed: {e}")
            return True
        else:
            return False

    console = Console()
    
    instructions = Text()
    instructions.append("Say 'Jarvis' followed by a command:\n\n", style="bold underline")
    instructions.append("- ", style="bold green")
    instructions.append("Open [Application]\n", style="cyan")
    instructions.append("- ", style="bold green")
    instructions.append("Google [Search Query]\n", style="cyan")
    instructions.append("- ", style="bold green")
    instructions.append("Youtube Search [Search Query]\n", style="cyan")
    instructions.append("- ", style="bold green")
    instructions.append("Create a new document\n", style="cyan")
    instructions.append("- ", style="bold green")
    instructions.append("Shutdown\n", style="cyan")
    instructions.append("- ", style="bold green")
    instructions.append("Or ask anything your AI assistant can help with— Jarvis can answer questions, explain concepts, and more!\n", style="bold cyan")
    panel = Panel(instructions, title="Jarvis Commands", border_style="blue")

    console.print(panel)

    def handle_stop_command():
        print("[Jarvis stopped speaking]")
        stop_event.set()
        with sentence_queue.mutex:
            sentence_queue.queue.clear()
        while state["is_speaking"]:
            time.sleep(0.05)
        stop_event.clear()
        state["is_stopped"] = False
        return True

    while True:
        partial_buffer = deque(maxlen=3)  # holds last few partial phrases

        while state["is_speaking"]:
            data = stream.read(4000, exception_on_overflow=False)
            partial = json.loads(recognizer.PartialResult()).get("partial", "").lower()
            if partial:
                partial_buffer.append(partial)

            combined = ' '.join(partial_buffer).strip()
            if "jarvis stop" in combined and not state["is_stopped"]:
                partial_buffer.clear()
                state["is_stopped"] = True
                print("[Detected 'Jarvis stop'] (during speech)")
                handle_stop_command()
                break


        data = stream.read(4000, exception_on_overflow=False)
        
        # partial = json.loads(recognizer.PartialResult()).get("partial", "").lower()
        # if "jarvis stop" in partial and not state["is_stopped"]:
        #         state["is_stopped"] = True
        #         print("[Detected 'Jarvis stop'] (while speaking)")
        #         handle_stop_command()
        #         continue

        if not state["is_speaking"] and recognizer.AcceptWaveform(data):
            result = json.loads(recognizer.Result())
            text = result.get("text", "").lower()

            if state["awaiting_shutdown"]:
                print("You said:", text)

                if text == "no":
                    print("Shutdown cancelled.")
                    stream.stop_stream()
                    yapr.yap("Shutdown cancelled.")
                    stream.start_stream()
                    state["awaiting_shutdown"] = False
                elif text == "yes":
                    print("JarvisOS is shutting down...")
                    stream.stop_stream()
                    yapr.yap("Shutting down. Goodbye.")
                    stream.start_stream()
                    sys.exit()
                else:
                    print("Sorry, I didn't understand. Would you like to shutdown?")
                    stream.stop_stream()
                    yapr.yap("Sorry, I didn't understand. Would you like to shutdown?")
                    stream.start_stream()
                continue

            # if "jarvis stop" in text:
            #     print("[Detected 'Jarvis stop']")
            #     handle_stop_command()
            
            elif "jarvis" in text:

                command = text.split("jarvis", 1)[-1].strip()

                # if "stop" in command:
                #     handle_stop_command()
                #     continue

                print("You said:", command)

                ret1 = google(command)
                ret2 = youtube(command)
                ret3 = open_software(command)
                ret4 = docs(command)
                ret5 = shutdown(command)

                if not (ret1 or ret2 or ret3 or ret4 or ret5) and command != "":
                    ask_lm_studio_streaming(command)

        # else:
        #     # ✅ Check partial results *here*, safely
        #     partial = json.loads(recognizer.PartialResult()).get("partial", "").lower()
        #     if "jarvis stop" in partial:
        #         print("[Detected 'Jarvis stop']")
        #         handle_stop_command()

            
            
            
if __name__ == "__main__":
    main()
        




import os
import subprocess
from abc import ABC, abstractmethod
from typing import Optional

import pyttsx3 as tts

import yapper.constants as c
from yapper.enums import PiperQuality, PiperVoice, PiperVoiceUS
from yapper.utils import (
    APP_DIR,
    download_piper_model,
    get_random_name,
    install_piper,
)

# suppresses pygame's welcome message
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"
import pygame  # noqa: E402


def play_wave(wave_f: str, volume: float = 1.0):
    """
    Plays the given wave file using pygame.

    Parameters
    ----------
    wave_f : str
        The wave file to play.
    volume : float, optional
        Volume to play the wave file between 0.0 and 1.0 (default: 1.0)
    """
    pygame.mixer.init()  # initialize pygame, safe to call multiple times
    sound = pygame.mixer.Sound(wave_f)
    sound.set_volume(volume)  # sets volume, defaults to 1

    sound.play()
    while pygame.mixer.get_busy():
        pygame.time.wait(100)


class BaseSpeaker(ABC):
    """
    Base class for speakers

    Methods
    ----------
    say(text: str)
        Speaks the given text.
    text_to_wave(text: str, file: str)
        Speaks the given text, saves the speech to a wave file.
    """

    @abstractmethod
    def say(self, text: str):
        pass

    @abstractmethod
    def text_to_wave(self, text: str, file: str):
        pass


class PyTTSXSpeaker(BaseSpeaker):
    """Converts speech to text using pyttsx."""

    def __init__(
        self,
        voice: str = c.VOICE_FEMALE,
        rate: int = c.SPEECH_RATE,
        volume: str = c.SPEECH_VOLUME,
    ):
        """
        Parameters
        ----------
        voice : str, optional
            Gender of the voice, can be 'f' or 'm' (default: 'f').
        rate : int, optional
            Rate of speech of the voice in wpm (default: 165).
        volume : float, optional
            Volume of the sound generated, can be 0-1 (default: 1).
        """
        assert voice in (
            c.VOICE_MALE,
            c.VOICE_FEMALE,
        ), "unknown voice requested"
        self.voice = voice
        self.rate = rate
        self.volume = volume

    def text_to_wave(self, text: str, file: str):
        """Saves the speech for the given text into the given file."""
        engine = tts.init()
        engine.setProperty("rate", self.rate)
        engine.setProperty("volume", self.volume)
        voice_id = engine.getProperty("voices")[
            int(self.voice == c.VOICE_FEMALE)
        ].id
        engine.setProperty("voice", voice_id)
        engine.save_to_file(text, file)
        engine.runAndWait()

    def say(self, text: str):
        """Speaks the given text"""
        engine = tts.init()
        engine.setProperty("rate", self.rate)
        engine.setProperty("volume", self.volume)
        voice_id = engine.getProperty("voices")[
            int(self.voice == c.VOICE_FEMALE)
        ].id
        engine.setProperty("voice", voice_id)
        engine.say(text)
        engine.runAndWait()


class PiperSpeaker(BaseSpeaker):
    """Converts text to speech using piper-tts"""

    def __init__(
        self,
        voice: PiperVoice = PiperVoiceUS.HFC_FEMALE,
        quality: Optional[PiperQuality] = None,
        show_progress: bool = True,
        volume: float = 1.0,
    ):
        """
        Parameters
        ----------
        voice : PiperVoice, optional
            Name of the piper voice to be used, can be one of PiperVoice*
            enums's attributes (default: PiperVoiceUS.AMY).
        quality : PiperQuality, optional
            Quality of the voice, can be ont of 'PiperQuality'
            enum's attributes (default: the highest available quality of
            the given voice).
        show_progress : bool
            Show progress when the voice model is being downloaded
            (default: True).
        volume : float, optional
            volume to play the wav file at, between 0.0 and 1.0
            (defaults to 1.0)
        """
        assert isinstance(
            voice, tuple(c.piper_enum_to_lang_code.keys())
        ), "voice must be a member of PiperVoice* enums"
        quality = quality or c.piper_voice_quality_map[voice]
        assert quality in PiperQuality, "quality must a member of PiperQuality"

        self.exe_path = str(install_piper(show_progress))
        self.onnx_f, self.conf_f = download_piper_model(
            voice, quality, show_progress
        )
        self.onnx_f, self.conf_f = str(self.onnx_f), str(self.conf_f)

        self.volume = volume

    def text_to_wave(self, text: str, file: str):
        """Saves the speech for the given text into the given file."""
        subprocess.run(
            [
                self.exe_path,
                "-m",
                self.onnx_f,
                "-c",
                self.conf_f,
                "-f",
                file,
                "-q",
            ],
            input=text.encode("utf-8"),
            stdout=subprocess.DEVNULL,
            check=True,
        )

    def say(self, text: str):
        """Speaks the given text"""
        f = APP_DIR / f"{get_random_name()}.wav"
        try:
            self.text_to_wave(text, str(f))
            play_wave(str(f), self.volume)
        finally:
            if f.exists():
                os.remove(f)



import threading
from functools import wraps
from types import TracebackType
from typing import Callable, Optional

import yapper.meta as meta
from yapper.enhancer import BaseEnhancer, NoEnhancer
from yapper.speaker import BaseSpeaker, PiperSpeaker


class Yapper:
    def __init__(
        self,
        enhancer: Optional[BaseEnhancer] = None,
        speaker: Optional[BaseSpeaker] = None,
        plain: bool = False,
        block: bool = True,
        use_stdout: bool = False,
    ):
        """
        Parameters
        ----------
        enhancer : Optional[BaseEnhancer]
            enhancer to be used for enhancing text
            (default: DefaultEnhancer).
        speaker : Optional[BaseSpeaker]
            speaker to be used for speaking the text
            (default: DefaultSpeaker).
        plain : bool, optional
            Do not enhance text, say it as it is (default: False).
        block: bool, optional
            wait for speech-syntheis to complete (default: True).
        use_stdout: bool, optional
            print the enhanced text before saying it (default: False).
        """
        self.enhancer = enhancer or NoEnhancer()
        self.speaker = speaker or PiperSpeaker()
        self.plain = plain
        self.block = block
        self.use_stdout = use_stdout

    def yap(
        self,
        text: str,
        plain: Optional[bool] = None,
        block: Optional[bool] = None,
        use_stdout: Optional[bool] = None,
    ):
        """
        Speaks the given text.

        Parameters
        ----------
        text : str
            The text to speak.
        plain : bool, optional
            Do not enhance text, say it as it is.
        block: Optional[bool]
            wait for speech-syntheis to complete.
        use_stdout: bool, optional
            print the enhanced text before saying it.
        """
        if plain is None:
            plain = self.plain
        if block is None:
            block = self.block
        if use_stdout is None:
            use_stdout = self.use_stdout

        def func(
            text: str,
            plain: bool,
            use_stdout: bool,
        ):
            if not plain:
                text = self.enhancer.enhance(text)
            if use_stdout:
                print(f"{meta.name}: {text}")
            self.speaker.say(text)

        if block:
            func(text, plain, use_stdout)
        else:
            threading.Thread(
                target=func, args=(text, plain, use_stdout)
            ).start()

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Exception],
        exc_value: Optional[str],
        tb: Optional[TracebackType],
    ):
        if exc_type:
            err = exc_type.__name__ + (f": {exc_value}" if exc_value else "")
            self.yap(err)
        return not exc_type

    def __call__(self, on_call: bool = False, pass_yapper: bool = False):
        """
        Parameters
        ----------
        on_call : bool, optional
            announce when the decorated function starts running
            (default: False).
        pass_yapper: bool, optional
            pass the yapper instance to the decorated function as a
            keyword argument 'yapper' when it is called (default: False).
        """

        def decorator(func: Callable):
            @wraps(func)
            def wrapper(*args, **kwargs):
                if on_call:
                    f_desc = f"running function {func.__name__}"
                    if func.__doc__ is not None:
                        f_desc += "\n" + f"description: {func.__doc__}"
                    self.yap(f_desc)
                try:
                    if not pass_yapper:
                        return func(*args, **kwargs)
                    return func(*args, **kwargs, yapper=self)
                except Exception as e:
                    err = e.__class__.__name__
                    desc = str(e)
                    if desc:
                        err = f"{err}: {desc}"
                    self.yap(err)
                    raise e

            return wrapper

        return decorator
